import errno
import logging
from collections import namedtuple

import galaxy.auth.providers
from galaxy.exceptions import Conflict
from galaxy.security.validate_user_input import validate_publicname
from galaxy.util import (
    parse_xml,
    parse_xml_string,
    plugin_config,
    string_as_bool,
)

log = logging.getLogger(__name__)

AUTH_CONF_XML = """<?xml version="1.0"?>
<auth>
    <authenticator>
        <type>localdb</type>
        <options>
            <allow-password-change>true</allow-password-change>
        </options>
    </authenticator>
</auth>
"""

Authenticator = namedtuple("Authenticator", ["plugin", "filter_template", "options"])


def get_authenticators(auth_config_file, auth_config_file_set):
    __plugins_dict = plugin_config.plugins_dict(galaxy.auth.providers, "plugin_type")
    # parse XML
    try:
        ct = parse_xml(auth_config_file)
        conf_root = ct.getroot()
    except OSError as exc:
        if exc.errno == errno.ENOENT and not auth_config_file_set:
            conf_root = parse_xml_string(AUTH_CONF_XML)
        else:
            raise

    authenticators = []
    # process authenticators
    for auth_elem in conf_root:
        type_elem_text = auth_elem.find("type").text
        plugin_class = __plugins_dict.get(type_elem_text)
        if not plugin_class:
            raise Exception(
                f"Authenticator type '{type_elem_text}' not recognized, should be one of {', '.join(__plugins_dict)}"
            )
        plugin = plugin_class()

        # check filterelem
        filter_elem = auth_elem.find("filter")
        if filter_elem is not None:
            filter_template = str(filter_elem.text)
        else:
            filter_template = None

        # extract options
        options_elem = auth_elem.find("options")
        options = {}
        if options_elem is not None:
            for opt in options_elem:
                options[opt.tag] = opt.text
        authenticator = Authenticator(
            plugin=plugin,
            filter_template=filter_template,
            options=options,
        )
        authenticators.append(authenticator)
    return authenticators


def parse_auth_results(trans, auth_results, options):
    auth_return = {}
    auth_result, auto_email, auto_username = auth_results[:3]
    auto_username = str(auto_username).lower()
    # make username unique
    max_retries = int(options.get("max-retries", "10"))
    try_number = 0
    while try_number <= max_retries:
        if try_number == 0:
            test_name = auto_username
        else:
            test_name = f"{auto_username}-{try_number}"
        validate_result = validate_publicname(trans, test_name)
        if validate_result == "":
            auto_username = test_name
            break
        else:
            log.debug(f"Invalid username '{auto_username}': {validate_result}")
        try_number += 1
    else:
        raise Conflict("Cannot make unique username")
    log.debug(f"Email: {auto_email}, auto-register with username: {auto_username}")
    auth_return["auto_reg"] = string_as_bool(options.get("auto-register", False))
    auth_return["email"] = auto_email
    auth_return["username"] = auto_username
    auth_return["auto_create_roles"] = string_as_bool(options.get("auto-create-roles", False))
    auth_return["auto_create_groups"] = string_as_bool(options.get("auto-create-groups", False))
    auth_return["auto_assign_roles_to_groups_only"] = string_as_bool(
        options.get("auto-assign-roles-to-groups-only", False)
    )

    if len(auth_results) == 4:
        auth_return["attributes"] = auth_results[3]
    return auth_return



"""
Contains implementations of the authentication logic.
"""

import logging

from galaxy.auth.util import (
    get_authenticators,
    parse_auth_results,
)
from galaxy.exceptions import Conflict
from galaxy.util import string_as_bool

log = logging.getLogger(__name__)


class AuthManager:
    def __init__(self, config):
        self.redact_username_in_logs = config.redact_username_in_logs
        self.authenticators = get_authenticators(config.auth_config_file, config.is_set("auth_config_file"))

    def check_registration_allowed(self, email, username, password, request):
        """Checks if the provided email/username is allowed to register."""
        message = ""
        status = "done"
        for provider, options in self.active_authenticators(email, username, password):
            allow_reg = _get_allow_register(options)
            if allow_reg == "challenge":
                auth_results = provider.authenticate(email, username, password, options, request)
                if auth_results[0] is True:
                    break
                if auth_results[0] is None:
                    message = "Invalid email address/username or password."
                    status = "error"
                    break
            elif allow_reg is True:
                break
            elif allow_reg is False:
                message = "Account registration not required for your account.  Please simply login."
                status = "error"
                break
        return message, status

    def check_auto_registration(self, trans, login, password, no_password_check=False):
        """
        Checks the username/email & password using auth providers in order.
        If a match is found, returns the 'auto-register' option for that provider.
        """
        if "@" in login:
            email = login
            username = None
        else:
            email = None
            username = login
        auth_return = {"auto_reg": False, "email": "", "username": ""}
        for provider, options in self.active_authenticators(email, username, password):
            if provider is None:
                log.debug(f"Unable to find module: {options}")
            else:
                options["no_password_check"] = no_password_check
                auth_results = provider.authenticate(email, username, password, options, trans.request)
                if auth_results[0] is True:
                    try:
                        auth_return = parse_auth_results(trans, auth_results, options)
                    except Conflict as conflict:
                        log.exception(conflict)
                        raise
                    return auth_return
                elif auth_results[0] is None:
                    log.debug("Login: '%s', stopping due to failed non-continue", login)
                    break  # end authentication (skip rest)
        return auth_return

    def check_password(self, user, password, request):
        """Checks the username/email and password using auth providers."""
        for provider, options in self.active_authenticators(user.email, user.username, password):
            if provider is None:
                log.debug(f"Unable to find module: {options}")
            else:
                auth_result = provider.authenticate_user(user, password, options, request)
                if auth_result is True:
                    return True  # accept user
                elif auth_result is None:
                    break  # end authentication (skip rest)
        return False

    def check_change_password(self, user, current_password, request):
        """Checks that auth provider allows password changes and current_password
        matches.
        """
        for provider, options in self.active_authenticators(user.email, user.username, current_password):
            if provider is None:
                log.debug(f"Unable to find module: {options}")
            else:
                auth_result = provider.authenticate_user(user, current_password, options, request)
                if auth_result is True:
                    if string_as_bool(options.get("allow-password-change", False)):
                        return
                    else:
                        return "Password change not supported."
                elif auth_result is None:
                    break  # end authentication (skip rest)
        return "Invalid current password."

    def active_authenticators(self, email, username, password):
        """Yields AuthProvider instances for the provided configfile that match the
        filters.
        """
        try:
            for authenticator in self.authenticators:
                filter_template = authenticator.filter_template
                if filter_template:
                    filter_str = filter_template.format(email=email, username=username, password=password)
                    passed_filter = eval(filter_str, {"__builtins__": None}, {"str": str})
                    if not passed_filter:
                        continue  # skip to next
                options = authenticator.options
                options["redact_username_in_logs"] = self.redact_username_in_logs
                yield authenticator.plugin, options
        except Exception:
            log.exception("Active Authenticators Failure")
            raise


def _get_allow_register(d):
    s = d.get("allow-register", True)
    if (lower_s := str(s).lower()) == "challenge":
        return lower_s
    else:
        return string_as_bool(s)
